#!/usr/bin/env python
# -*- coding: utf-8 -*-

# A SBB/CFF/FFS commandline based timetable client.
# Copyright (C) 2012-2019 Danilo Bargen
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

import sys
import logging
import argparse

from . import meta
from .parser import parse_input
from .api import get_connections
from .display import Formats, connectionsTable
from .helpers import perror


def main():
    output_format = Formats.SIMPLE
    proxy_host = None

    # 1. Parse command line arguments
    parser = argparse.ArgumentParser(epilog='Arguments:\n'
                + ' You can use natural language arguments using the following\n'
                + ' keywords in your desired language:\n'
                + ' en -- from, to, via, departure, arrival\n'
                + ' de -- von, nach, via, ab, an\n'
                + ' fr -- de, à, via, départ, arrivée\n'
                + '\n'
                + ' You can also use natural time and date specifications in your language, like:\n'
                + ' - "now", "immediately", "at noon", "at midnight",\n'
                + ' - "tomorrow", "monday", "in 2 days", "22/11".\n'
                + '\n'
                + 'Examples:\n'
                + ' fahrplan from thun to burgdorf\n'
                + ' fahrplan via bern nach basel von zürich, helvetiaplatz ab 15:35\n'
                + ' fahrplan de lausanne à vevey arrivée minuit\n'
                + ' fahrplan from Bern to Zurich departure 13:00 monday\n'
                + ' fahrplan -p proxy.mydomain.ch:8080 de lausanne à vevey arrivée minuit\n'
                + '\n', formatter_class=argparse.RawDescriptionHelpFormatter, prog=meta.title, description=meta.description, add_help=False)
    parser.add_argument("--full", "-f", action="store_true", help="Show full connection info, including changes")
    parser.add_argument("--info", "-i", action="store_true", help="Verbose output")
    parser.add_argument("--debug", "-d", action="store_true", help="Debug output")
    parser.add_argument("--help", "-h", action="store_true", help="Show this help")
    parser.add_argument("--version", "-v", action="store_true", help="Show version number")
    parser.add_argument("--proxy", "-p", help="Use proxy for network connections (host:port)")
    parser.add_argument("request", nargs=argparse.REMAINDER)
    options = parser.parse_args()

    # Version
    if options.version:
        print('{meta.title} {meta.version}'.format(meta=meta))
        sys.exit(0)

    # No request or help
    if len(options.request) == 0 or options.help:
        parser.print_help()
        sys.exit(0)

    # Options
    if options.full:
        output_format = Formats.FULL
    if options.debug:
        logging.basicConfig(level=logging.DEBUG)
    if options.proxy is not None:
        proxy_host = options.proxy

    # Parse user request
    try:
        args, language = parse_input(options.request)
    except ValueError as e:
        perror('Error:', e)
        sys.exit(1)

    # 2. API request
    data = get_connections(args, (output_format == Formats.FULL), proxy_host)
    connections = data["connections"]

    if not connections:
        print("No connections found")
        sys.exit(0)

    # 3. Output data
    table = connectionsTable(connections, output_format)
    print(table)

if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
from texttable import Texttable


# Output formats
class Formats(object):
    SIMPLE = 0
    FULL = 1


def _get_connection_row(i, connection):
    """
    Get table row for connection.
    """
    sections = connection['sections']
    # Create row
    row = [i]
    for p in [
            lambda x: [x['station_from'], x['station_to']],  # Station
            lambda x: [x.get('platform_from') or '-', x.get('platform_to') or '-'],  # Platform
            lambda x: [x[q].strftime('%d/%m/%y') for q in ['departure', 'arrival']],  # Date
            lambda x: [x[q].strftime('%H:%M') for q in ['departure', 'arrival']],  # Time
            lambda x: [str((x['arrival'] - x['departure'])).rsplit(':', 1)[0], ' '],  # Duration
            None,
            lambda x: [x['travelwith'], ' '],  # With
            lambda x: ['1: ' + x['occupancy1st'], '2: ' + x['occupancy2nd']]  # Occupancy
    ]:
        if p is None:
            row.append(connection['change_count'])
        else:
            inner = ['\n'.join(p(s)) for s in sections]
            row.append('\n \n'.join(inner))
    return row


def connectionsTable(connections, output_format):
    """
    Get connections in the given output format.
    """
    table = Texttable(max_width=0)
    # Alignments
    table.set_cols_valign(['m', 't', 't', 't', 't', 't', 'm', 't', 't'])
    table.set_cols_align(['l', 'l', 'c', 'l', 'l', 'c', 'c', 'l', 'l'])
    # Header
    table.add_row(['#', 'Station', 'Platform', 'Date', 'Time', 'Duration', 'Chg.', 'With', 'Occupancy'])
    # Connection rows
    for i, connection in enumerate(connections):
        table.add_row(_get_connection_row(i, connection))
    # Display
    return table.draw()



"""This module contains metadata about the project."""
title = 'fahrplan'
version = '1.1.0'
description = 'A SBB/CFF/FFS commandline based timetable client.'
author = 'Danilo Bargen and contributors'
author_email = 'mail@dbrgn.ch'
url = 'https://github.com/dbrgn/fahrplan'
keywords = 'fahrplan timetable sbb cff ffs public transport'
license = 'GPLv3'



# -*- coding: utf-8 -*-
import requests
import logging
import json
import dateutil.parser
import sys
from .helpers import perror

occupancies = {
    None: '',
    -1: '',
    0: 'Low',  # todo check
    1: 'Low',
    2: 'Medium',
    3: 'High',
}
API_URL = 'http://transport.opendata.ch/v1'


def _api_request(action, params, proxy=None):
    """
    Perform an API request on transport.opendata.ch
    """
    # Send request
    url = "{}/{}".format(API_URL, action)
    kwargs = {'params': params}
    if proxy is not None:
        kwargs['proxies'] = {'http': proxy}
    try:
        response = requests.get(url, **kwargs)
    except requests.exceptions.ConnectionError:
        perror('Error: Could not reach network.')
        sys.exit(1)

    # Check response status
    logging.debug('Response status: {0!r}'.format(response.status_code))
    if not response.ok:
        verbose_status = requests.status_codes._codes[response.status_code][0]
        perror('Server Error: HTTP {} ({})'.format(response.status_code, verbose_status))
        sys.exit(1)

    # Convert response to json
    try:
        return json.loads(response.text)
    except ValueError:
        logging.debug('Response status code: {0}'.format(response.status_code))
        logging.debug('Response content: {0!r}'.format(response.content))
        perror('Error: Invalid API response (invalid JSON)')
        sys.exit(1)


def _parse_section(con_section, connection):
    """
    Parse the section of a connection
    """
    departure = con_section['departure']
    arrival = con_section['arrival']
    journey = con_section.get('journey')
    walk = con_section.get('walk')
    section = {}
    section['station_from'] = departure['station']['name']
    section['station_to'] = arrival['station']['name']
    section['travelwith'] = journey["name"] if journey is not None else ""
    section['departure'] = dateutil.parser.parse(departure['departure'])
    section['arrival'] = dateutil.parser.parse(arrival['arrival'])
    section['platform_from'] = "" if walk else departure['platform']
    section['platform_to'] = arrival['platform']
    if walk:
        section['occupancy1st'] = ''
        section['occupancy2nd'] = ''
    elif journey:
        section['occupancy1st'] = occupancies.get(con_section['journey']['capacity1st'], '')
        section['occupancy2nd'] = occupancies.get(con_section['journey']['capacity2nd'], '')
    else:
        section['occupancy1st'] = occupancies.get(connection['capacity1st'], '')
        section['occupancy2nd'] = occupancies.get(connection['capacity2nd'], '')
    return section


def _parse_connection(connection, include_sections=False):
    """Parse a connection.

    Process a connection object as returned from the API and return a
    dictionary with cleaned data.

    Args:
        connection: A connection dictionary as returned by the JSON API.
        include_sections: A boolean value determining whether to include the
            sections in the returned data set or not (default False).

    Returns:
        A dictionary containing cleaned data. If sections are enabled, they are
        contained in a list.
    """
    data = {}
    walk = False

    def keyfunc(s):
        return s['departure']['departure']
    data['change_count'] = str(connection['transfers'])
    data['travelwith'] = ', '.join(connection['products'])
    data['occupancy1st'] = occupancies.get(connection['capacity1st'], '')
    data['occupancy2nd'] = occupancies.get(connection['capacity2nd'], '')

    # Sections
    con_sections = sorted(connection['sections'], key=keyfunc)
    data['sections'] = []
    if include_sections:
        # Full display
        for con_section in con_sections:
            section = _parse_section(con_section, connection)
            if con_section.get('walk'):
                walk = True
            data["sections"].append(section)
    else:
        # Shortened display, parse only departure and arrivals sections
        section = _parse_section(con_sections[0], connection)
        to = _parse_section(con_sections[-1], connection)
        for p in ["station_to", "arrival"]:
            section[p] = to[p]
        # Get information from connection
        for p in ["occupancy2nd", "occupancy1st", "travelwith", "change_count"]:
            section[p] = data[p]
        data['sections'] = [section]

    # Walk
    if walk:
        data['travelwith'] += ', Walk'

    return data


def get_connections(request, include_sections=False, proxy=None):
    """
    Get the connections of a request
    """
    data = _api_request("connections", request, proxy)
    data["connections"] = [_parse_connection(c, include_sections) for c in data["connections"]]
    return data



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from setuptools import setup
from fahrplan import meta

f = open('requirements.txt', 'r')
lines = f.readlines()
requirements = [l.strip().strip('\n') for l in lines if l.strip() and not l.strip().startswith('#')]
readme = open('README.rst').read()

setup(name='fahrplan',
      version=meta.version,
      description=meta.description,
      author=meta.author,
      author_email=meta.author_email,
      url=meta.url,
      packages=['fahrplan'],
      zip_safe=False,
      include_package_data=True,
      license=meta.license,
      keywords=meta.keywords,
      long_description=readme,
      install_requires=requirements,
      entry_points={
          'console_scripts': [
              '%s = fahrplan.main:main' % meta.title,
          ]
      },
      classifiers=[
          'Development Status :: 5 - Production/Stable',
          'Environment :: Console',
          'License :: OSI Approved :: GNU General Public License (GPL)',
          'Natural Language :: English',
          'Natural Language :: German',
          'Natural Language :: French',
          'Natural Language :: Italian',
          'Operating System :: MacOS',
          'Operating System :: POSIX :: Linux',
          'Programming Language :: Python :: 3',
          'Programming Language :: Python :: 3.5',
          'Programming Language :: Python :: 3.6',
          'Programming Language :: Python :: 3.7',
          'Programming Language :: Python :: 3.8',
          'Topic :: Internet',
          'Topic :: Terminals',
      ],
)



# -*- coding: utf-8 -*-
from __future__ import print_function, division, absolute_import, unicode_literals

import sys
from datetime import datetime

from subprocess import Popen, PIPE
if sys.version_info[0] == 2 and sys.version_info[1] < 7:
    import unittest2 as unittest
else:
    import unittest

from .. import meta
from .. import parser


BASE_COMMAND = 'python -m fahrplan.main'
try:
    ENCODING = sys.stdout.encoding or 'utf-8'
except AttributeError:
    ENCODING = 'utf-8'


# Run command
class CommandOutput(object):
    def __init__(self, stdout, stderr, code):
        self.std_err = stderr
        self.std_out = stdout
        self.status_code = code


def run_command(command):
    p = Popen(command, stdout=PIPE, stderr=PIPE, shell=True)
    stdout, stderr = p.communicate()
    return CommandOutput(stdout.decode(ENCODING), stderr.decode(ENCODING), p.returncode)


class TestBasicArgumentHandling(unittest.TestCase):

    def testRequiredArgumentsMissing(self):
        r = run_command('{0} von bern'.format(BASE_COMMAND))
        self.assertEqual('Error: "from" and "to" arguments must be present!\n', r.std_err)

    def testVersionInfo(self):
        args = ['-v', '--version', '-d -i -v']
        for arg in args:
            r = run_command('{0} {1}'.format(BASE_COMMAND, arg))
            self.assertEqual('%s %s\n' % (meta.title, meta.version), r.std_out)

    def testHelp(self):
        args = ['-h', '--help', '-d -i --help']
        for arg in args:
            r = run_command('{0} {1}'.format(BASE_COMMAND, arg))
            self.assertTrue(meta.description in r.std_out)
            self.assertTrue('usage:' in r.std_out)
            self.assertTrue('optional arguments:' in r.std_out)
            self.assertTrue('positional arguments:' in r.std_out)
            self.assertTrue('Examples:' in r.std_out)


class TestInputParsing(unittest.TestCase):

    valid_expected_result = {
        'arrival': '19:00',
        'departure': '18:30',
        'from': 'Zürich',
        'to': 'Locarno',
        'via': 'Genève',
    }

    def testEmptyArguments(self):
        tokens = []
        data, language = parser.parse_input(tokens)
        self.assertEqual({}, data)
        self.assertIsNone(language)

    def testOneValidArgument(self):
        tokens = 'from'.split()
        data, language = parser.parse_input(tokens)
        self.assertEqual({}, data)
        self.assertIsNone(language)

    def testOneInvalidArgument(self):
        tokens = 'foobar'.split()
        data, language = parser.parse_input(tokens)
        self.assertEqual({}, data)
        self.assertIsNone(language)

    def testValidArgumentsEn(self):
        tokens = 'from Zürich to Locarno via Genève departure 18:30 arrival 19:00'.split()
        data, language = parser._process_tokens(tokens, sloppy_validation=True)
        self.assertEqual(self.valid_expected_result, data)
        self.assertEqual('en', language)

    def testValidArgumentsDe(self):
        tokens = 'von Zürich nach Locarno via Genève ab 18:30 an 19:00'.split()
        data, language = parser._process_tokens(tokens, sloppy_validation=True)
        self.assertEqual(self.valid_expected_result, data)
        self.assertEqual('de', language)

    def testValidArgumentsFr(self):
        tokens = 'de Zürich à Locarno via Genève départ 18:30 arrivée 19:00'.split()
        data, language = parser._process_tokens(tokens, sloppy_validation=True)
        self.assertEqual(self.valid_expected_result, data)
        self.assertEqual('fr', language)

    def testTwoArguments(self):
        tokens = 'Zürich Basel'.split()
        data, language = parser.parse_input(tokens)
        self.assertEqual({'from': 'Zürich', 'to': 'Basel'}, data)
        self.assertEqual('en', language)

    def testNotEnoughArgument(self):
        tokens = 'from basel via bern'.split()
        self.assertRaises(ValueError, parser.parse_input, tokens)

    def testBasicDepartureTime(self):
        tokens = 'von basel nach bern ab 1945'.split()
        expected = {'from': 'basel', 'time': '19:45', 'to': 'bern'}
        self.assertEqual(expected, parser.parse_input(tokens)[0])

    def testBasicArrivalTime(self):
        tokens = 'von basel nach bern an 18:00'.split()
        expected = {'from': 'basel', 'isArrivalTime': 1, 'time': '18:00', 'to': 'bern'}
        self.assertEqual(expected, parser.parse_input(tokens)[0])

    def testImmediateTimes(self):
        now = datetime.now().strftime('%H:%M')
        queries = [
            'von basel nach bern ab jetzt'.split(),
            'von basel nach bern ab sofort'.split(),
            'from basel to bern departure now'.split(),
            'from basel to bern departure right now'.split(),
            'from basel to bern departure immediately'.split(),
            'de basel à bern départ maitenant'.split(),
        ]
        for tokens in queries:
            data, _ = parser.parse_input(tokens)
            self.assertEqual(now, data['time'])

    def testNoonTimes(self):
        queries = [
            'von basel nach bern ab mittag'.split(),
            'from basel to bern departure noon'.split(),
            'de basel à bern départ midi'.split(),
        ]
        for tokens in queries:
            data, _ = parser.parse_input(tokens)
            self.assertEqual('12:00', data['time'])

    def testMidnightTimes(self):
        queries = [
            'von basel nach bern ab mitternacht'.split(),
            'from basel to bern departure midnight'.split(),
            'de basel à bern départ minuit'.split(),
        ]
        for tokens in queries:
            data, _ = parser.parse_input(tokens)
            self.assertEqual('23:59', data['time'])

    def testAtTimes(self):
        queries = [
            'von basel nach bern ab am mittag'.split(),
            'von basel nach bern ab um 12:00'.split(),
            'from basel to bern departure at noon'.split(),
            'from basel to bern departure at 12:00'.split(),
        ]
        for tokens in queries:
            data, _ = parser.parse_input(tokens)
            self.assertEqual('12:00', data['time'])

    def testDates(self):
        year = datetime.now().year
        queries = [
            'von basel nach bern ab 22/10/{} 13:00'.format(year).split(),
            'von basel nach bern ab um 22/10 13:00'.split(),
            'from basel to bern departure 22/10 13:00'.split(),
            'from basel to bern departure 22/10 13:00'.split(),
        ]
        for tokens in queries:
            data, _ = parser.parse_input(tokens)
            self.assertEqual('13:00', data['time'])
            self.assertEqual('{}/10/22'.format(year), data['date'])


class TestBasicQuery(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        """Setup method that is only run once."""
        cmd = '{0} von basel nach zürich ab 07:00'.format(BASE_COMMAND)
        cls.r = run_command(cmd)
        cls.rows = cls.r.std_out.split('\n')

    def returnStatus(self):
        """The command should return the status code 0."""
        self.assertEqual(0, self.r.status_code)

    def testRowCount(self):
        """A normal output table should have 16 rows."""
        self.assertEqual(16, len(self.rows))

    def testHeadline(self):
        """Test the headline items."""
        headline_items = ['Station', 'Platform', 'Date', 'Time',
                          'Duration', 'Chg.', 'With', 'Occupancy']
        for item in headline_items:
            self.assertIn(item, self.rows[1])

    def testStationNames(self):
        """Station names should be "Basel SBB" and "Zürich HB"."""
        self.assertTrue("Basel SBB" in self.rows[3])
        self.assertTrue("Zürich HB" in self.rows[4])


class TestLanguages(unittest.TestCase):

    def testBasicQuery(self):
        """
        Test a query in three languages and assert that the output of all
        three queries is equal.

        """
        args = ['von bern nach basel via zürich ab 15:00',
                'from bern to basel via zürich departure 15:00',
                'de bern à basel via zürich départ 15:00']
        jobs = [run_command('{0} {1}'.format(BASE_COMMAND, arg)) for arg in args]

        statuscodes = [job.status_code for job in jobs]
        self.assertEqual([0, 0, 0], statuscodes)

        stdout_values = [job.std_out for job in jobs]
        self.assertTrue(stdout_values[1:] == stdout_values[:-1])


class RegressionTests(unittest.TestCase):

    def testIss11(self):
        """
        Github issue #11:
        Don't allow both departure and arrival time.
        """
        args = 'von bern nach basel ab 15:00 an 16:00'
        query = run_command('{0} {1}'.format(BASE_COMMAND, args))
        self.assertEqual('Error: You can\'t specify both departure *and* arrival time.\n',
                query.std_err)

    def testIss13(self):
        """
        Github issue #13:
        Station not found: ValueError: max() arg is an empty sequence.
        """
        args = 'von zuerich manegg nach nach stadelhofen'
        query = run_command('{0} {1}'.format(BASE_COMMAND, args))
        self.assertEqual(0, query.status_code, 'Program terminated with statuscode != 0')


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
from datetime import datetime, timedelta
import re
import logging

keywords = {
    'de': {
        'now': ['jetzt', 'sofort', 'nun'],
        'noon': ['mittag'],
        'midnight': ['mitternacht'],
        'today': ["heute"],
        'tomorrow': ["morgen"],
        'at': ['um', 'am'],
        'days': [r'in (\d+) tagen'],
        'weekdays': ["montag", "dienstag", "mittwoch", "donnerstag", "freitag", "samstag", "sonntag"],
    },
    'en': {
        'now': ['now', 'right now', 'immediately'],
        'noon': ['noon'],
        'midnight': ['midnight'],
        'today': ["today"],
        'tomorrow': ["tomorrow"],
        'at': ['at'],
        'days': [r'in (\d+) days'],
        'weekdays': ["monday", "tuesday", "wednesday", "thursday", "friday", "saturday", "sunday"],
    },
    'fr': {
        'now': ['maitenant'],
        'noon': ['midi'],
        'midnight': ['minuit'],
        'today': ["aujourd'hui"],
        'tomorrow': ["demain"],
        'days': [r"dans (\d+) jours"],
        'at': [],  # TODO: "à" clashes with top level keywords
        'weekdays': ["lundi", "mardi", "mercredi", "jeudi", "vendredi", "samedi", "dimanche"],
    },
}


def _process_tokens(tokens, sloppy_validation=False):
    """Parse input tokens.

    Take a list of tokens (usually ``sys.argv[1:]``) and parse the "human
    readable" input into a format suitable for machines.

    Args:
        tokens: List of tokens (usually ``sys.argv[1:]``.
        sloppy_validation: Set to True to enable less strict validation. Used
            mainly for testing, default False.

    Returns:
        A 2-tuple containing the unmapped data dictionary and the language
        string. For example:

        ({'to': 'bern', 'from': 'zürich', 'departure': '18:00'}, 'de')

    Raises:
        ValueError: If "from" or "to" arguments are missing or if both
            departure *and* arrival time are specified (as long as
            sloppy_validation is disabled).

    """
    if len(tokens) < 2:
        return {}, None

    keyword_dicts = {
        'en': {'from': 'from', 'to': 'to', 'via': 'via',
               'departure': 'departure', 'arrival': 'arrival'},
        'de': {'from': 'von', 'to': 'nach', 'via': 'via',
               'departure': 'ab', 'arrival': 'an'},
        'fr': {'from': 'de', 'to': 'à', 'via': 'via',
               'departure': 'départ', 'arrival': 'arrivée'},
    }

    # Detect language
    language = _detect_language(keyword_dicts, tokens)
    logging.info('Detected [%s] input' % language)

    # Keywords mapping
    keywords = dict((v, k) for k, v in keyword_dicts[language].items())
    logging.debug('Using keywords: ' + ', '.join(keywords.keys()))

    # Prepare variables
    data = {}
    stack = []

    def process_stack():
        """Process the stack. First item is the key, rest is value."""
        key = keywords.get(stack[0])
        value = ' '.join(stack[1:])
        data[key] = value
        stack[:] = []

    # Process tokens
    for token in tokens:
        if token in keywords.keys():
            if stack:
                process_stack()
        elif not stack:
            continue
        stack.append(token)

    if not stack:
        return {}, None

    process_stack()

    # Validate data
    if not sloppy_validation:
        if not ('from' in data and 'to' in data):
            raise ValueError('"from" and "to" arguments must be present!')
        if 'departure' in data and 'arrival' in data:
            raise ValueError('You can\'t specify both departure *and* arrival time.')

    return data, language


def _detect_language(keyword_dicts, tokens):
    """Detect the language of the tokens by finding the highest intersection
    with the keywords of a specific language."""

    def intersection_count(a, b):
        return len(set(a).intersection(b))

    counts = []
    for lang, keywords in keyword_dicts.items():
        count = intersection_count(keywords.values(), tokens)
        counts.append((lang, count))

    language = max(counts, key=lambda x: x[1])[0]
    return language


def _parse_date(datestring, keywords):
    """Parse date tokens.

    Args:
        datestring: String containing a date specification.
        keywords: Language keywords

    Returns:
        date string.

    Raises:
        ValueError: If time could not be parsed.

    """
    date = None
    days_shift = None
    # Keywords
    for i, d in enumerate(["today", "tomorrow"]):
        if any([t in datestring for t in keywords[d]]):
            days_shift = i
    # Weekdays
    for i, d in enumerate(keywords["weekdays"]):
        if d in datestring:
            days_shift = i - datetime.now().weekday()
            if days_shift <= 0:
                days_shift += 7
    # Shifts
    if days_shift is None:
        for pattern in keywords["days"]:
            days_re = re.search(pattern, datestring)
            if days_re:
                try:
                    days_shift = int(days_re.group(1))
                except:
                    pass

    if days_shift is not None:
        return datetime.now() + timedelta(days=days_shift)

    # Regular date strings
    for dateformat in [[r"(\d{2}/\d{2}/\d{4})", "%d/%m/%Y"], [r"(\d{2}/\d{2})", "%d/%m"]]:
        days_re = re.search(dateformat[0], datestring)
        if days_re:
            try:
                date = datetime.strptime(days_re.group(1), dateformat[1])
            except:
                continue
            if date.year == 1900:
                date = date.replace(year=datetime.now().year)
            break

    if date is not None:
        return date.strftime("%Y/%m/%d")

    return None


def _parse_time(timestring, keywords):
    """Parse time tokens.

    Args:
        timestring: String containing a time specification.
        keywords: Language keywords

    Returns:
        Time string.

    Raises:
        ValueError: If time could not be parsed.

    """

    # Ignore "at" keywords
    if timestring.split(' ', 1)[0] in keywords['at']:
        timestring = timestring.split(' ', 1)[1]

    # Parse regular time strings
    # regular_time_match = re.search(r'([0-2]?[0-9])[:\-\. ]([0-9]{2})', timestring)
    regular_time_match = re.search(r'(?<!/)(\d{2})(?::*)(\d{2})', timestring)
    if regular_time_match:
        return ':'.join(regular_time_match.groups())
    timestring = timestring.lower()
    if timestring in keywords['now']:
        return datetime.now().strftime('%H:%M')
    if timestring in keywords['noon']:
        return '12:00'
    if timestring in keywords['midnight']:
        return '23:59'  # '00:00' would be the first minute of the day, not the last one.
    raise ValueError('Time is missing or could not be parsed')


def parse_input(tokens):
    """Parse input tokens.

    Take a list of tokens (usually ``sys.argv[1:]``) and parse the "human
    readable" input into a format suitable for machines. The output format
    matches the format required by the Transport API.

    Args:
        tokens: List of tokens (usually ``sys.argv[1:]``.

    Returns:
        A 2-tuple containing the data dictionary and the language string. For
        example:

        ({'to': 'bern', 'from': 'zürich'}, 'de')

    Raises:
        ValueError: If "from" or "to" arguments are missing or if both
            departure *and* arrival time are specified.

    """

    # Process tokens, get data dict and language
    data, language = _process_tokens(tokens)
    if data == {}:
        if len(tokens) == 2:
            # Fallback when e.g. doing "fahrplan zurich basel"
            return {'from': tokens[0], 'to': tokens[1]}, 'en'
        return data, language

    try:
        kws = keywords[language]
    except IndexError:
        raise ValueError('Invalid language: "%s"!' % language)
    # Map keys
    for t in ["departure", "arrival"]:
        if t in data:
            data["time"] = _parse_time(data[t], kws)
            date = _parse_date(data[t], kws)
            if date is not None:
                data["date"] = date
            if t == "arrival":
                data['isArrivalTime'] = 1
            del data[t]

    logging.debug('Data: ' + repr(data))
    return data, language
